///////////////////////////////////////////////////////////////////////////////
// BSD 3-Clause License
//
// Copyright (C) 2019-2020, LAAS-CNRS, University of Edinburgh
// Copyright note valid unless otherwise stated in individual files.
// All rights reserved.
///////////////////////////////////////////////////////////////////////////////

#ifndef CROCODDYL_MULTIBODY_COST_BASE_HPP_
#define CROCODDYL_MULTIBODY_COST_BASE_HPP_

#include "crocoddyl/core/utils/deprecate.hpp"
CROCODDYL_PRAGMA_DEPRECATED_HEADER("crocoddyl/multibody/cost-base.hpp",
                                   "crocoddyl/core/cost-base.hpp")
#include "crocoddyl/core/cost-base.hpp"

#endif
